% Code for Digitizing the Data

%% File reading

fname = 'XXX.tif'; % Write the Raw TIF Image File Name

info = imfinfo(fname);   % imfinfo : information of the image file
num_images = numel(info); % Number of Images in tif file

for k = 1:num_images
    A(:,:,k) = imread(fname, k, 'Info', info); % Read TIF into 3D array A
end

%% Specify Position of Optical Tweezer Trap

numtweezers = 198; % number of tweezers

load('xxyy_init.mat'); % Load the Initially Identified Tweezer Position; xx is the x position and yy is the y position

% Because the x position and y position of the trap might slightly change in long time due to thermal noise, mechanical imperfections, etc., we
% re-identify the position of trap with following lines.

% We start from the initial position, regard it as a center, find the maximum flouresence point within (numwidth) x (numwidth) (in this case 5x5)

numofwidth=11;
numwidth=5;

for k=1:length(xx)
    % Find the maximum average fluoresence for all experiment repetitions
    [i,j]=ind2sub([numofwidth,numofwidth],find((mean(A(yy(k)-numwidth:yy(k)+numwidth,xx(k)-numwidth:xx(k)+numwidth,:),3))...
        ==max(max(mean(A(yy(k)-numwidth:yy(k)+numwidth,xx(k)-numwidth:xx(k)+numwidth,:),3)))==1)); 
    i=i(1); j=j(1);
    yy(k) = yy(k)-numwidth-1+i;
    xx(k) = xx(k)-numwidth-1+j;
end

% See whether the xx yy positions are well identified; can be commented
% figure; imagesc(mean(A,3)); set(gca,'YDir','normal');
% hold on; scatter(xx,yy,'r')

%% fluoA : Fluorescence value of each trap position for all repetition

% fluoA is 2D array; its dimension is (numtweezers, num_images)

numofwidth=5;
numwidth=2;

fluoA = zeros(numtweezers,num_images);  %% tweezer number x number of images

for k=1:numtweezers
    
    % We average average photon count of k-th tweezer (flouresence due to atom)
    % Averaging (numwidth) x (numwidth) square where center is (xx(k),yy(k)), for our case 2x2 square
    
    fluoA(k,:,:,:,:) = sum(sum(A(yy(k)-numwidth:yy(k)+numwidth,xx(k)-numwidth:xx(k)+numwidth,:),1),2)/(numofwidth^2);
    
end

%% threshold : Find the threshold of fluorescence from the tweezers

% Fluorescence of tweezers is background noise when there is no atom, and get higher value in average when there is an atom inside the tweezer
% Therefore by make the histogram of flouresence we could discriminate whether atom exist or not in the tweezer. 

threshold = zeros(numtweezers,1);

for k = 1:numtweezers
    figure(100+floor((k-1)/6/9)); % Each figure draw 6*9 = 54 histograms; this is arbitrary
    subplot(6,9,k-floor((k-1)/6/9)*54); % Draw subplot
    title(num2str(k)); % Subplot name is atom #
    
    h=histogram(fluoA(k,:));
    threstmp = mean(fluoA(k,:));
    h.BinEdges = [threstmp-45,threstmp+75]; % Total range of histogram; arbitrary and could be modified
    h.NumBins = 70; % Total bins of histogram; arbitrary and could be modified
    
    % following code is for finding the valley between two photon counts
    % valley + some arbitrary value is regarded as the threshold for distinguishing existence of atom
    
    [va,ind] = min(h.Values(10:25));
    threshold(k) = h.BinEdges(ind+9);
    hold on;
    plot(threshold(k),10,'d');

    drawnow;
end

%% fluodig : 1 if greater than threshold, 0 if less than threshold

% fluodig is 2D array; its dimension is (numtweezers, num_images)

fluodig = zeros(numtweezers,num_images);

% Digitize the flouresence data according to threshold

for k=1:length(threshold)
    fluodig(k,:)=(sign(fluoA(k,:)-threshold(k))+1)/2;
end

fluodig(fluodig==1/2)=1;

%% fluoAreshape, fluodigreshape: Reshape flouA, fluodig into 3D array

numexp = floor(length(fluodig)/2); % Total experiment (repetition) number. The number 2 is because we take two photos for every experiment.

% flouAreshape and fluodigreshape is 3D array; its dimension is (numtweezers, 2 (because of two photographs), numexp)

fluoAreshape = reshape(fluoA(:,1:2*numexp),numtweezers, 2, numexp);

fluodigreshape = reshape(fluodig(:,1:2*numexp),numtweezers, 2, numexp);
