#!/usr/bin/env python
# coding: utf-8

# In[1]:

# This code makes the chains of the nodes, each containing unique graph and its property

# GraphLinkedList is a linked list of the 'GraphNode's, where they are linked with the GraphNode property next and prev

# Each GraphNode consists of key (Weisfeiler-Lehman graph hash), exactresult (exact MIS solution), expresult (list of experimented MIS solutions), graph (keyG from GraphTable), next (next GraphNode), prev (previous GraphNode)


# Reference
# https://seongonion.tistory.com/

import networkx as nx
import numpy as np

class GraphNode:
    def __init__(self, key, exactresult, expresult, graph): # 각 노드가 key, value를 갖는다
        self.key = key
        self.exactresult = exactresult
        self.expresult = expresult
        self.graph = graph
        self.next = None
        self.prev = None
        

# GraphLinkedList Class for Chaining Graph Nodes

class GraphLinkedList:
    def __init__(self):
        self.head = None
        self.tail = None
    
    def append(self, key, exactresult, expresult, graph): # Append new node at the end of the chain of nodes
        new_node = GraphNode(key, exactresult, expresult, graph)
        if self.head == None:
            self.head = new_node
            self.tail = new_node
        else:
            self.tail.next = new_node
            new_node.prev = self.tail
            self.tail = new_node
        
    def find_node_with_key(self, key): # Check whether there exist a node with the key (hash of graph)
        iterator = self.head
        
        while iterator:
            if iterator.key == key:
                return iterator
                
            iterator = iterator.next

        return None
        
    def delete(node_to_delete):
        # Delete last remaining node
        if node_to_delete is self.head and node_to_delete is self.tail:
            self.tail = None
            self.head = None

        # Delete first node
        elif node_to_delete is self.head:
            self.head = self.head.next
            self.head.prev = None

        # Delete last node
        elif node_to_delete is self.tail:
            self.tail = self.tail.prev
            self.tail.next = None

        # Delete other cases
        else:
            node_to_delete.prev.next = node_to_delete.next
            node_to_delete.next.prev = node_to_delete.prev

        return node_to_delete.value
        
        
    def __str__(self):
        res_str = ""
        iterator = self.head
        while iterator:
            res_str += "{}: exact {} : meanexp {} : expnum {}\n".format(iterator.key, iterator.exactresult, np.array(iterator.expresult).mean(), len(iterator.expresult))
            iterator = iterator.next
            
        return res_str

