#!/usr/bin/env python
# coding: utf-8

# In[10]:

# This code makes the table of graphs

# GraphTable is made of python list with 198 entries (maximum atom number)

# Each entries represent the size of the graph |G| which is being stacked. 
# e.g. The index 8 of the list represents the graphs of |G|=8+1=9

# Each entries are filled with the linked list class named 'GraphLinkedList', which is in separate python script

# References
# https://seongonion.tistory.com/
# https://stackoverflow.com/questions/20530455/isomorphic-comparison-of-networkx-graph-objects-instead-of-the-default-addres


import GraphLinkedList as LL
import networkx as nx
import MISSolver as MISSolver

class GraphTable:
    def __init__(self): # Create GraphTable instance as list of the GraphLinkedList with capacity of total atom number
        self.capacity = 198 # total atom number
        self.table = [LL.GraphLinkedList() for _ in range(self.capacity)] # generate GraphLinkedList with capacitiy of total atom number
        
    def hash_function(self, keyG):
        return nx.weisfeiler_lehman_graph_hash(keyG) # Use Weisfeiler-Lehman graph hash for hashing the graph;
                                                     # We determine graph isomorphism by comparing this hash.
                                                     # This hash function gurantees that isomorphic graphs have same hash;
                                                     # More over, it strongly guarantees that non-isomorphic graphs have different hash.
                                                     # Nino, et al. Journal of Machine Learning Research 12.9 (2011).
                    
    def get_linked_list(self, numatom): # Find the linked list corresponding to numatom        
        return self.table[numatom-1]
    
    def look_up_node(self, numatom, keyG): # Find whether GraphTable has node with the networkx graph object 'keyG'
        linked_list = self.get_linked_list(numatom)
        return linked_list.find_node_with_key(self.hash_function(keyG)) # Using method from 'GraphLinkedList'
        
    def insert(self, numatom, keyG, expresult): # Insert new node in GraphTable
    
        existing_node = self.look_up_node(numatom, keyG)
        
        if existing_node: # If keyG exist append experiment result
            existing_node.expresult.append(expresult)
            
        else: # If keyG doesn't exist add new node
            linked_list = self.get_linked_list(numatom)
            missol = MISSolver.MISSolver() # instance of MISSolver class
            exactresulttot = missol.mis2(keyG) # Solve MIS of the graph keyG; the method mis(keyG) returns size of MIS and the MIS itself
            exactresult = exactresulttot[0] # Get the size of exactly solved MIS size
            linked_list.append(self.hash_function(keyG), exactresult, [expresult], keyG) # Using method from 'GraphLinkedList', append hash, exactsol, 
                                                                                         # experimented solution list, and graph keyG
            
    def __str__(self):

        res_str = ""

        for linked_list in self.table:
            res_str += str(linked_list)

        return res_str[:-1]

