#!/usr/bin/env python
# coding: utf-8

# In[1]:

# https://github.com/GiggleLiu/MISAlgorithms.jl

import networkx as nx
import numpy as np
import copy
import matplotlib.pylab as plt
class MISSolver:
    def __init__(self):
        pass
        #print('successfully created,',__name__)

    def mis2_every_solutions(self,eg):
        def NeighborCover(eg,vi): # Get neighbors of vertex `vi`, including itself.
            return list(nx.all_neighbors(eg,vi))+[vi]
        Nodes = list(eg.nodes)
        MIS_num_candidates = []
        MIS_nodes_candidates = []
        for node in Nodes:
            
            #Get mis solution with and without node.
            eg_A = copy.deepcopy(eg)
            eg_A.remove_nodes_from(NeighborCover(eg,node))
            eg_B = copy.deepcopy(eg)
            eg_B.remove_node(node)
            A_result = self.mis2(eg_A) #with node
            MIS_num_A = A_result[0]+1; MIS_set_A = A_result[1].union(set([node]))
            B_result = self.mis2(eg_B) #without node
            MIS_num_B = B_result[0]; MIS_set_B = B_result[1]
            if MIS_num_A >= MIS_num_B:
                MIS_num_candidates.append(MIS_num_A)
                MIS_nodes_candidates.append(MIS_set_A)
            else:
                MIS_num_candidates.append(MIS_num_B)
                MIS_nodes_candidates.append(MIS_set_B)
        MIS_num = np.max(MIS_num_candidates)
        MIS_nodes_list = []
        for i, item in enumerate(MIS_num_candidates):
            if item == MIS_num:
                mayput = set(MIS_nodes_candidates[i])
                if mayput not in MIS_nodes_list:
                    MIS_nodes_list.append(mayput)
        return MIS_num, MIS_nodes_list
            
    def mis2(self,eg):
        def isconnected(eg,a,b):
            if (a,b) in eg.edges(data=False) or (b,a) in eg.edges(data=False):
                return 1
            else:
                return 0
        def mindegree_vertex(eg):
            node_and_degree = list(eg.degree(list(eg.nodes)))
            degree = []
            for item in node_and_degree:
                degree.append(item[1])
            return node_and_degree[np.argmin(degree)]
        def maxdegree_vertex(eg):
            node_and_degree = list(eg.degree(list(eg.nodes)))
            degree = []
            for _ in range(len(eg.nodes)):
                degree.append(node_and_degree[1])
            return node_and_degree[np.argmax(degree)]
        def eliminate(func,eg,vi):
            eg_copy = copy.deepcopy(eg)
            if type(vi) == int or type(vi) == str or type(vi) == np.int64:
                eg_copy.remove_node(vi)
                res=func(eg_copy)
            else:# vi is given in array or list.
                eg_copy.remove_nodes_from(vi)
                res=func(eg_copy)
            return res
        def NeighborCover(eg,vi): # Get neighbors of vertex `vi`, including itself.
            return list(nx.all_neighbors(eg,vi))+[vi]
        def neighbors(eg,vmin):
            return list(nx.all_neighbors(eg,vmin))
        def neighbors2(eg,v):
            #Get 2nd nearest neighbors, i.e. distance 2 vertices.
            Nv_list = neighbors(eg,v)
            N2v_set = set([])
            for w in Nv_list:
                Nw_list = neighbors(eg,w)
                for x in Nw_list:
                    if x != v:
                        N2v_set.add(x)
            return list(N2v_set)
        def mirror(eg,v): 
            # Get mirrors of vertex `vi`, as well as itself.
            #A vertex `w` is a mirror of `v` if `w ∈ N²(v)` and `N[v]-N(w)` is a clique.
            N2v_list = neighbors2(eg,v)
            Nv_list = neighbors(eg,v)
            return_list = []
            for w in N2v_list:
                Nw_list = neighbors(eg,w)
                Nv_minus_Nw_list = [item for item in Nv_list if item not in Nw_list]
                H = eg.subgraph(Nv_minus_Nw_list)
                n=len(Nv_minus_Nw_list)
                if len(H.edges) == n*(n-1)/2:
                    return_list.append(w)
            return return_list
        def MirrorCover(eg,v):
            return mirror(eg,v)+[v]
        def find_cluster(eg,v):
            #Find the cluster connected to `v` in `eg`.
            def FC(eg,v,vset):
                if v not in vset:
                    vset.add(v)
                    for vj in neighbors(eg,v):
                        FC(eg,vj,vset)
                else:
                    return
            vset = set([])
            for node in eg.nodes:
                if len(neighbors(eg,node)) == 0: vset.add(node)
            FC(v,vset)
            return list(vset)
        def adjacent45(eg):
            #find adjacent vertice with degree 4, 5.
            node_and_degree_view = eg.degree(list(eg.nodes))
            for vi in eg.nodes:
                deg = node_and_degree_view[vi]
                for vj in neighbors(eg,vi):
                    if node_and_degree_view[vj] != deg: return (vi,vj)
                
                
        # eg: nx.Graph instance
        if len(eg.nodes) == 0:
            return 0,set([])
        elif len(eg.nodes) == 1:
            return 1,set(eg.nodes)
        elif len(eg.nodes) == 2:
            if isconnected(eg,list(eg.nodes)[0],list(eg.nodes)[1]):
                return 1,set([list(eg.nodes)[0]]) # selecting frist node for convenience...
            else:
                return 2,set(eg.nodes) # two isolated nodes!
        elif len(eg.nodes) == 3:
            a,b,c=list(eg.nodes)[0:3]
            nedge = len(eg.edges)
            if nedge == 0: return 3,set(eg.nodes)
            elif nedge == 3: return 1,set([list(eg.nodes)[0]])# selecting frist node for convenience...
            else: 
                if not isconnected(eg,list(eg.nodes)[0],list(eg.nodes)[1]):
                    return 2,set([list(eg.nodes)[0],list(eg.nodes)[1]])
                elif not isconnected(eg,list(eg.nodes)[0],list(eg.nodes)[2]):
                    return 2,set([list(eg.nodes)[0],list(eg.nodes)[2]])
                else:
                    return 2,set([list(eg.nodes)[1],list(eg.nodes)[2]])
        else:
            vmin,degmin = mindegree_vertex(eg)
            if degmin == 0: # DONE
                eliminated = eliminate(self.mis2,eg,vmin)
                eliminated[1].add(vmin)
                return 1 + eliminated[0],eliminated[1]
            elif degmin == 1: # DONE
                eliminated = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                eliminated[1].add(vmin)
                return 1 + eliminated[0],eliminated[1]
            elif degmin == 2:  #DONE
                a,b = neighbors(eg,vmin)
                if isconnected(eg,a,b):
                    eliminated = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                    eliminated[1].add(vmin)
                    return 1+eliminated[0],eliminated[1]
                else:
                    sn=neighbors2(eg,vmin)
                if len(sn) == 1:
                    w=sn[0]
                    eliminated = eliminate(self.mis2,eg,[vmin,w,a,b])
                    eliminated[1].add(a); eliminated[1].add(b) # need revision if erroneous
                    return 2+eliminated[0], eliminated[1]
                else:
                    eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                    eliminatedB = eliminate(self.mis2,eg,MirrorCover(eg,vmin))
                    candidates = [1+eliminatedA[0],eliminatedB[0]]
                    if np.argmax(candidates) == 0:
                        eliminatedA[1].add(vmin)
                        return candidates[0], eliminatedA[1]
                    else:
                        return candidates[1], eliminatedB[1]
            elif degmin == 3: #DONE
                a,b,c = neighbors(eg,vmin)
                nedge = isconnected(eg,a,b)+isconnected(eg,a,c)+isconnected(eg,b,c)
                if nedge == 0:
                    ms = MirrorCover(eg,vmin)
                    if len(ms) > 1:
                        eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                        eliminatedB = eliminate(self.mis2,eg,ms)
                        candidates = [1+eliminatedA[0],eliminatedB[0]]
                        if np.argmax(candidates) == 0:
                            eliminatedA[1].add(vmin)
                            return candidates[0], eliminatedA[1]
                        else:
                            return candidates[1], eliminatedB[1]
                    else:
                        eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                        eliminatedB = eliminate(self.mis2,eg,list(set(NeighborCover(eg,a)).union(set(NeighborCover(eg,b)))))
                        eliminatedC = eliminate(self.mis2,eg,list(set(NeighborCover(eg,a)).union(set(NeighborCover(eg,c))).union(set([b]))))
                        eliminatedD = eliminate(self.mis2,eg,list(set(NeighborCover(eg,b)).union(set(NeighborCover(eg,c))).union(set([a]))))
                        candidates = [1+eliminatedA[0],2+eliminatedB[0],2+eliminatedC[0],2+eliminatedD[0]]
                        argmax = np.argmax(candidates)
                        if argmax == 0:
                            eliminatedA[1].add(vmin)
                            return candidates[argmax],eliminatedA[1]
                        elif argmax == 1:
                            eliminatedB[1].add(a);eliminatedB[1].add(b)
                            return candidates[argmax],eliminatedB[1]
                        elif argmax == 2:
                            eliminatedC[1].add(a);eliminatedC[1].add(c)
                            return candidates[argmax],eliminatedC[1]
                        elif argmax == 3:
                            eliminatedD[1].add(b);eliminatedD[1].add(c)
                            return candidates[argmax],eliminatedD[1]
                elif nedge == 3:
                    eliminated = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                    eliminated[1].add(vmin)
                    return 1+eliminated[0],eliminated[1]
                else:
                    eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,vmin))
                    eliminatedB = eliminate(self.mis2,eg,MirrorCover(eg,vmin))
                    candidates = [1+eliminatedA[0],eliminatedB[0]]
                    argmax = np.argmax(candidates)
                    if argmax == 0:
                        eliminatedA[1].add(vmin)
                        return candidates[argmax],eliminatedA[1]
                    else:
                        return candidates[argmax],eliminatedB[1]
            else:
                vmax,degmax = maxdegree_vertex(eg)
                if degmax >= 6:
                    eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,vmax))
                    eliminatedB = eliminate(self.mis2,eg,vmax)
                    candidates = [1+eliminatedA[0],eliminatedB[0]]
                    argmax = np.argmax(candidates)
                    if argmax == 0:
                        eliminatedA[1].add(vmax)
                        return candidates[argmax],eliminatedA[1]
                    else:
                        return candidates[argmax],eliminatedB[1]
                elif nx.is_connected(eg) == False:
                    cluster = find_cluster(eg,vmax)
                    A =eg.subgraph(cluster)
                    cluster_c = [node for node in eg.nodes if node not in cluster]
                    B =eg.subgraph(cluster_c)
                    
                    mis2A = mis2(A)
                    mis2B = mis2(B)
                    
                    return mis2A[0]+mis2B[0],mis2A[1].union(mis2B[1])
                elif degmin == degmax:
                    eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,vmax))
                    eliminatedB = eliminate(self.mis2,eg,MirrorCover(eg,vmax))
                    candidates = [1+eliminatedA[0],eliminatedB[0]]
                    argmax = np.argmax(candidates)
                    if argmax == 0:
                        eliminatedA[1].add(vmax)
                        return candidates[argmax],eliminatedA[1]
                    else:
                        return candidates[argmax],eliminatedB[1]
                else:
                    v4,v5= adjacent45(eg)
                    
                    eliminatedA = eliminate(self.mis2,eg,NeighborCover(eg,v5))
                    eliminatedB = eliminate(self.mis2,eg,list(set(MirrorCover(eg,v5)).union(set(NeighborCover(eg,v4)))))
                    eliminatedC = eliminate(self.mis2,eg,list(set(MirrorCover(eg,v5)).union(set([v4]))))
                    candidates=[1+eliminatedA[0],1+eliminatedB[0],eliminatedC[0]]
                    argmax = np.argmax(candidates)
                    if argmax == 0:
                        eliminatedA[1].add(v5)
                        return candidates[argmax],eliminatedA[1]
                    elif argmax == 1:
                        eliminatedB[1].add(v4)
                        return candidates[argmax],eliminatedB[1]
                    elif argmax == 2:
                        return candidates[argmax],eliminatedC[1]

